import logging
from typing import List

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.logging import _log_melodi_http_errors
from melodi.user_internal_for_project.data_models import (
    BulkUserInternalForProjectRequest, BulkUserInternalForProjectResponse)


class UserInternalForProjectClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = (
            self.base_url + "/api/external/user-internal-for-project/bulk"
        )
        self.endpoint = (
            self.base_endpoint + f"?apiKey={self.api_key}"
        )

        self.logger = logging.getLogger(__name__)

    def set_users_internal(self, project_id: int, user_ids: List[int]) -> BulkUserInternalForProjectResponse:
        url = self.endpoint

        request = BulkUserInternalForProjectRequest(
            projectId=project_id,
            externalUserIds=user_ids
        )

        try:
            response = requests.request("POST", url, json=request.dict(by_alias=True))

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()

            return parse_obj_as(BulkUserInternalForProjectResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def set_users_not_internal(self, project_id: int, user_ids: List[int]) -> None:
        url = self.endpoint

        request = BulkUserInternalForProjectRequest(
            projectId=project_id,
            externalUserIds=user_ids
        )

        try:
            response = requests.request("DELETE", url, json=request.dict(by_alias=True))

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()

            return None
        except MelodiAPIError as e:
            raise MelodiAPIError(e)



import unittest

from melodi.messages.data_models import Message
from melodi.utils.openai_prompt_parser import (
    _process_prompt_message,
    _extract_chat_prompt,
    _get_generation_metadata,
    _get_melodi_messages_from_openai_prompt,
)
from melodi.utils.openai_utils import OpenAiDefinition


class TestOpenAIPromptParserTests(unittest.TestCase):
    def test_process_prompt_message(self):
        self.assertIsNone(_process_prompt_message(None))
        self.assertEqual(_process_prompt_message("test"), "test")
        self.assertEqual(_process_prompt_message({}), {})
        self.assertEqual(
            _process_prompt_message({"content": "test"}), {"content": "test"}
        )
        self.assertEqual(
            _process_prompt_message({"content": "test"}), {"content": "test"}
        )
        self.assertEqual(
            _process_prompt_message(
                {
                    "content": [
                        {"type": "input_audio", "message": "should be ignored"},
                        {"type": "not_audio", "message": "should be added"},
                    ]
                }
            ),
            {
                "content": [
                    {"type": "not_audio", "message": "should be added"},
                ]
            },
        )

    def test_extract_chat_prompt(self):
        self.assertEqual(_extract_chat_prompt({}), {"messages": []})
        self.assertEqual(_extract_chat_prompt({"messages": []}), {"messages": []})
        self.assertEqual(
            _extract_chat_prompt(
                {
                    "messages": [
                        {
                            "content": [
                                {"type": "not_audio", "message": "should be added"},
                            ]
                        }
                    ]
                }
            ),
            {
                "messages": [
                    {
                        "content": [
                            {"type": "not_audio", "message": "should be added"},
                        ]
                    }
                ]
            },
        )
        self.assertEqual(
            _extract_chat_prompt({"messages": ["hello"]}),
            {"messages": ["hello"]},
        )
        # TODO add unit tests for function_calls, functions, tools

    def test_get_generation_metadata(self):
        self.assertEqual(
            _get_generation_metadata({}),
            {
                "frequency_penalty": 0,
                "logprobs": False,
                "n": 1,
                "parallel_tool_calls": True,
                "presence_penalty": 0,
                "reasoning_effort": "medium",
                "service_tier": "auto",
                "store": False,
                "stream": False,
                "temperature": 1,
                "top_p": 1,
            },
        )

        self.assertEqual(
            _get_generation_metadata({"logprobs": True, "other_key": False, "n": 4}),
            {
                "frequency_penalty": 0,
                "logprobs": True,
                "n": 4,
                "parallel_tool_calls": True,
                "presence_penalty": 0,
                "reasoning_effort": "medium",
                "service_tier": "auto",
                "store": False,
                "stream": False,
                "temperature": 1,
                "top_p": 1,
            },
        )

    def test_get_melodi_messages_from_openai_prompt(self):
        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={},
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="completion",
                    sync=True,
                ),
            ),
            [],
        )
        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={},
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="test",
                    sync=True,
                ),
            ),
            [],
        )
        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={"prompt": {}},
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="completion",
                    sync=True,
                ),
            ),
            [],
        )

        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={
                    "prompt": {
                        "messages": [{"role": "test_role", "content": "Hello World"}]
                    }
                },
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="completion",
                    sync=True,
                ),
            ),
            [
                Message(
                    externalId=f"input_0",
                    role="Test_Role",
                    content="Hello World",
                    metadata={
                        "type": "input_message",
                        "frequency_penalty": 0,
                        "logprobs": 0,
                        "n": 1,
                        "parallel_tool_calls": 1,
                        "presence_penalty": 0,
                        "reasoning_effort": "medium",
                        "service_tier": "auto",
                        "store": 0,
                        "stream": 0,
                        "temperature": 1,
                        "top_p": 1,
                    },
                )
            ],
        ),

        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={
                    "prompt": {
                        "messages": [{"role": "test_role", "content": "Hello World"}, {"role": "system", "content": "Hello World Back"}]
                    },
                    "temperature": 4
                },
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="completion",
                    sync=True,
                ),
            ),
            [
                Message(
                    externalId=f"input_0",
                    role="Test_Role",
                    content="Hello World",
                    metadata={
                        "type": "input_message",
                        "frequency_penalty": 0,
                        "logprobs": 0,
                        "n": 1,
                        "parallel_tool_calls": 1,
                        "presence_penalty": 0,
                        "reasoning_effort": "medium",
                        "service_tier": "auto",
                        "store": 0,
                        "stream": 0,
                        "temperature": 4,
                        "top_p": 1,
                    },
                ),
                Message(
                    externalId=f"input_1",
                    role="System",
                    content="Hello World Back",
                    metadata={
                        "type": "input_message",
                        "frequency_penalty": 0,
                        "logprobs": 0,
                        "n": 1,
                        "parallel_tool_calls": 1,
                        "presence_penalty": 0,
                        "reasoning_effort": "medium",
                        "service_tier": "auto",
                        "store": 0,
                        "stream": 0,
                        "temperature": 4,
                        "top_p": 1,
                    },
                )
            ],
        )

        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={
                    "prompt": {
                        "messages": [{"role": "test_role", "content": "Hello World"},
                                     {"role": "system", "content": "Hello World Back"}]
                    },
                    "temperature": 4
                },
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="chat",
                    sync=True,
                ),
            ),
            [],
        )

        self.assertEqual(
            _get_melodi_messages_from_openai_prompt(
                kwargs={
                    "messages": [
                        {"role": "test_role", "content": "Hello World"},
                        {"role": "system", "content": "Hello World Back"}
                    ],
                    "temperature": 4,
                    "frequency_penalty": 2,
                },
                openai_resource=OpenAiDefinition(
                    module="openai",
                    object="Completion",
                    method="create",
                    type="chat",
                    sync=True,
                ),
            ),
            [
                Message(
                    externalId=f"input_0",
                    role="Test_Role",
                    content="Hello World",
                    metadata={
                        "type": "input_message",
                        "frequency_penalty": 2,
                        "logprobs": 0,
                        "n": 1,
                        "parallel_tool_calls": 1,
                        "presence_penalty": 0,
                        "reasoning_effort": "medium",
                        "service_tier": "auto",
                        "store": 0,
                        "stream": 0,
                        "temperature": 4,
                        "top_p": 1,
                    },
                ),
                Message(
                    externalId=f"input_1",
                    role="System",
                    content="Hello World Back",
                    metadata={
                        "type": "input_message",
                        "frequency_penalty": 2,
                        "logprobs": 0,
                        "n": 1,
                        "parallel_tool_calls": 1,
                        "presence_penalty": 0,
                        "reasoning_effort": "medium",
                        "service_tier": "auto",
                        "store": 0,
                        "stream": 0,
                        "temperature": 4,
                        "top_p": 1,
                    },
                )
            ],
        )


if __name__ == "__main__":
    unittest.main()



import json
import logging
from dataclasses import dataclass
from typing import Optional

import types
import openai.resources
from packaging.version import Version

from datetime import datetime, timezone


logger = logging.getLogger(__name__)


class OpenAiKwargsExtractor:
    def __init__(
        self,
        **kwargs,
    ):
        self.kwargs = kwargs

    def get_openai_args(self):
        return self.kwargs


@dataclass
class OpenAiDefinition:
    module: str
    object: str
    method: str
    type: str
    sync: bool
    min_version: Optional[str] = None


OPENAI_CLIENTS_V0 = [
    OpenAiDefinition(
        module="openai",
        object="ChatCompletion",
        method="create",
        type="chat",
        sync=True,
    ),
    OpenAiDefinition(
        module="openai",
        object="Completion",
        method="create",
        type="completion",
        sync=True,
    ),
]

OPENAI_CLIENTS_V1 = [
    OpenAiDefinition(
        module="openai.resources.chat.completions",
        object="Completions",
        method="create",
        type="chat",
        sync=True,
    ),
    OpenAiDefinition(
        module="openai.resources.completions",
        object="Completions",
        method="create",
        type="completion",
        sync=True,
    ),
    OpenAiDefinition(
        module="openai.resources.chat.completions",
        object="AsyncCompletions",
        method="create",
        type="chat",
        sync=False,
    ),
    OpenAiDefinition(
        module="openai.resources.completions",
        object="AsyncCompletions",
        method="create",
        type="completion",
        sync=False,
    ),
]

NON_STREAM_MESSAGE_KEYS = [
    "created",
    "id",
    "model",
    "object",
    "service_tier",
    "system_fingerprint",
]

COMPLETION_CHOICE_KEYS = [
    "finish_reason",
    "index",
    "logprobs",
]

COMPLETION_CHOICE_MESSAGE_KEYS = [
    "content",
    "refusal",
    "role",
    "annotations",
    "function_call",
    "tool_calls",
]

COMPLETION_USAGE_KEYS = [
    "total_tokens",
]

COMPLETION_USAGE_TOKENS_KEYS = [
    "accepted_prediction_tokens",
    "reasoning_tokens",
    "rejected_prediction_tokens",
]

COMPLETION_USAGE_PROMPT_TOKENS_KEYS = [
    "cached_tokens",
]

GENERATION_PARAMETERS_DEFAULTS = {
    "frequency_penalty": 0,
    "logit_bias": None,
    "logprobs": False,
    "max_completion_tokens": None,
    "max_tokens": None,
    "modalities": None,
    "n": 1,
    "parallel_tool_calls": True,
    "prediction": None,
    "presence_penalty": 0,
    "reasoning_effort": "medium",
    "response_format": None,
    "seed": None,
    "service_tier": "auto",
    "stop": None,
    "store": False,
    "stream": False,
    "stream_options": None,
    "temperature": 1,
    "tool_choice": None,
    "top_logprobs": None,
    "top_p": 1,
    "user": None,
    "web_search_options": None,
    "metadata": None,
}


def time_now(user_timezone=timezone.utc, as_string: bool = False):
    if as_string:
        return str(datetime.now(user_timezone))

    return datetime.now(user_timezone)


def _is_openai_v1():
    return Version(openai.__version__) >= Version("1.0.0")


def _is_streaming_response(response):
    return (
        isinstance(response, types.GeneratorType)
        or isinstance(response, types.AsyncGeneratorType)
        or (_is_openai_v1() and isinstance(response, openai.Stream))
        or (_is_openai_v1() and isinstance(response, openai.AsyncStream))
    )


def clean_dict_value(d):
    if not isinstance(d, dict):
        return d if d else False  # use a custom test if needed
    return {key: v for key, value in d.items() if (v := clean_dict_value(value))}


def parse_metadata_value(input_value):
    input_value = strip_openai_objects(input_value)

    if isinstance(input_value, int) or isinstance(input_value, str):
        return input_value

    if isinstance(input_value, float):
        return str(input_value)

    if isinstance(input_value, list) or isinstance(input_value, dict):
        return json.dumps(input_value)

    logger.info(f"Could not parse metadata value: {input_value}")
    return None


def strip_openai_objects(input_value):
    if (
        input_value is None
        or isinstance(input_value, int)
        or isinstance(input_value, str)
        or isinstance(input_value, float)
    ):
        return input_value

    if isinstance(input_value, list):
        return [strip_openai_objects(element) for element in input_value]

    if isinstance(input_value, dict):
        return {key: strip_openai_objects(value) for key, value in input_value.items()}

    input_value = input_value.__dict__
    return {key: strip_openai_objects(value) for key, value in input_value.items()}



import logging
from typing import List

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.logging import _log_melodi_http_errors
from melodi.user_segment_types.data_models import (UserSegmentTypeDefinition,
                                                   UserSegmentTypesQueryParams)


class UserSegmentTypesClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = (
            self.base_url + "/api/external/users/segment-types"
        )
        self.endpoint = (
            self.base_endpoint + f"?apiKey={self.api_key}"
        )

        self.logger = logging.getLogger(__name__)

    def get(self, query_params: UserSegmentTypesQueryParams = UserSegmentTypesQueryParams()) -> List[UserSegmentTypeDefinition]:
        url = self.endpoint

        if (query_params.projectId):
            url = f"{url}&projectId={query_params.projectId}"

        try:
            response = requests.request("GET", url)

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()

            return parse_obj_as(List[UserSegmentTypeDefinition], response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)



import logging

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.feedback.data_models import (Feedback,
                                         FeedbackCreateOrUpdateRequest,
                                         FeedbackResponse)
from melodi.logging import _log_melodi_http_errors


class FeedbackClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = (
            self.base_url + "/api/external/feedback"
        )
        self.endpoint = (
            self.base_endpoint + f"?apiKey={self.api_key}"
        )

        self.logger = logging.getLogger(__name__)

    def create(self, feedback: Feedback) -> FeedbackResponse:
        try:
            createdAtString = None
            if (feedback.createdAt):
                createdAtString = feedback.createdAt.isoformat()
            feedbackJson = feedback.dict(by_alias=True)
            feedbackJson['createdAt'] = createdAtString

            response = requests.request(
                "POST",
                url=self.endpoint,
                json=feedbackJson,
                headers=self._get_headers(),
            )

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(FeedbackResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def create_or_update(self, update: FeedbackCreateOrUpdateRequest) -> FeedbackResponse:
        try:
            createdAtString = None
            if (update.createdAt):
                createdAtString = update.createdAt.isoformat()
            feedbackJson = update.dict(by_alias=True)
            feedbackJson['createdAt'] = createdAtString

            response = requests.request(
                "PUT",
                url=self.endpoint,
                json=feedbackJson,
                headers=self._get_headers(),
            )

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(FeedbackResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)




from openai import NotGiven

from melodi.messages.data_models import Message
from melodi.utils.openai_utils import (
    GENERATION_PARAMETERS_DEFAULTS,
    parse_metadata_value,
    OpenAiDefinition,
)
from melodi.utils.utils import handle_melodi_failure


@handle_melodi_failure("Could not parse OpenAI prompt attributes")
def _get_melodi_messages_from_openai_prompt(
    kwargs: dict, openai_resource: OpenAiDefinition
):
    """Convert the OpenAI prompt in Melodi messages."""
    if openai_resource.type == "completion":
        prompt = kwargs.get("prompt", {})
    elif openai_resource.type == "chat":
        prompt = _extract_chat_prompt(kwargs)
    else:
        # Currently only chat and completion objects are supported
        return []

    if not prompt.get("messages"):
        return []

    generation_metadata = _get_generation_metadata(kwargs)

    prompt_messages = prompt.pop("messages")
    message_metadata = {
        "type": "input_message",
    }
    message_metadata.update(**prompt)
    message_metadata.update(**generation_metadata)

    melodi_messages = []
    for message in prompt_messages:
        melodi_messages.append(
            Message(
                externalId=f"input_{len(melodi_messages)}",
                role=message["role"].title(),
                content=message["content"],
                metadata=message_metadata,
            )
        )

    return melodi_messages


def _get_generation_metadata(kwargs: dict):
    """Accessed 3/31/25: https://platform.openai.com/docs/api-reference/chat/create"""
    generation_metadata = dict()
    for key_param, key_default_value in GENERATION_PARAMETERS_DEFAULTS.items():
        value = kwargs.get(key_param)

        if value is None or isinstance(value, NotGiven):
            if key_default_value is not None:
                generation_metadata[key_param] = key_default_value
        else:
            generation_metadata[key_param] = parse_metadata_value(value)

    return generation_metadata


def _extract_chat_prompt(kwargs: dict):
    """Return prompt messages and potential function calls for the prompt."""
    prompt = dict()

    for key in ["function_call", "functions", "tools"]:
        if not kwargs.get(key):
            continue

        prompt[key] = parse_metadata_value(kwargs.get(key))

    prompt_messages = [
        _process_prompt_message(message) for message in kwargs.get("messages", [])
    ]
    prompt["messages"] = prompt_messages
    return prompt


def _process_prompt_message(message):
    if not isinstance(message, dict):
        return message

    processed_message = {**message}

    content = processed_message.get("content")
    if not isinstance(content, list):
        return processed_message

    processed_content = []

    for content_part in content:
        if content_part.get("type") == "input_audio":
            # Not supported
            pass
        else:
            processed_content.append(content_part)

    processed_message["content"] = processed_content

    return processed_message



import logging

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.logging import _log_melodi_http_errors
from melodi.threads.data_models import (Thread, ThreadResponse,
                                        ThreadsPagedResponse,
                                        ThreadsQueryParams)


class ThreadsClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = self.base_url + "/api/external/threads"
        self.endpoint = self.base_endpoint + f"?apiKey={self.api_key}"

        self.logger = logging.getLogger(__name__)


    def create(self, thread: Thread) -> ThreadResponse:
        createdAtString = None
        if (thread.createdAt):
            createdAtString = thread.createdAt.isoformat()
        threadjson = thread.dict(by_alias=True)
        threadjson['createdAt'] = createdAtString

        try:
            response = requests.post(
                self.endpoint, headers=self._get_headers(), json=threadjson
            )
            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(ThreadResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def create_or_update(self, thread: Thread) -> ThreadResponse:
        createdAtString = None
        if (thread.createdAt):
            createdAtString = thread.createdAt.isoformat()
        threadjson = thread.dict(by_alias=True)
        threadjson['createdAt'] = createdAtString

        try:
            response = requests.put(
                self.endpoint, headers=self._get_headers(), json=threadjson
            )
            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(ThreadResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def get(self, query_params: ThreadsQueryParams = ThreadsQueryParams()) -> ThreadsPagedResponse:
        url = f"{self.endpoint}&pageIndex={query_params.pageIndex}&pageSize={query_params.pageSize}"

        if (query_params.projectId):
            url = f"{url}&projectId={query_params.projectId}"
        if (query_params.ids):
            for threadId in query_params.ids:
                url = f"{url}&ids={threadId}"
        if (query_params.externalIds):
            for externalId in query_params.externalIds:
                url = f"{url}&externalIds={externalId}"
        if (query_params.before):
            url = f"{url}&before={query_params.before.isoformat()}"
        if (query_params.after):
            url = f"{url}&after={query_params.after.isoformat()}"
        if (query_params.search):
            url = f"{url}&search={query_params.search}"
        if (query_params.userSegmentIds):
            for userSegmentId in query_params.userSegmentIds:
                url = f"{url}&userSegmentIds={userSegmentId}"
        if (query_params.issueIds):
            for issueId in query_params.issueIds:
                url = f"{url}&issueIds={issueId}"
        if (query_params.intentIds):
            for intentId in query_params.intentIds:
                url = f"{url}&intentId={intentId}"
        if (query_params.hasFeedback):
            url = f"{url}&hasFeedback={query_params.hasFeedback}"
        if (query_params.feedbackType):
            url = f"{url}&feedbackType={query_params.feedbackType}"
        if (query_params.attributeOptionIds):
            for attributeOptionId in query_params.attributeOptionIds:
                url = f"{url}&attributeOptionIds={attributeOptionId}"
        if (query_params.includeFeedback):
            url = f"{url}&includeFeedback={query_params.includeFeedback}"
        if (query_params.includeIntents):
            url = f"{url}&includeIntents={query_params.includeIntents}"
        if (query_params.includeIssues):
            url = f"{url}&includeIssues={query_params.includeIssues}"
        if (query_params.outcome):
            url = f"{url}&outcome={query_params.outcome}"
        if (query_params.filterInternal):
            url = f"{url}&filterInternal={query_params.filterInternal}"
        if (query_params.filterAnonymousSessions):
            url = f"{url}&filterAnonymousSessions={query_params.filterAnonymousSessions}"
        if (query_params.metadataField):
            url = f"{url}&metadataField={query_params.metadataField}"
        if (query_params.metadataValue):
            url = f"{url}&metadataValue={query_params.metadataValue}"


        try:
            response = requests.request("GET", url)

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()

            return parse_obj_as(ThreadsPagedResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)



from datetime import datetime
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Extra

from melodi.messages.data_models import Message, MessageResponse
from melodi.users.data_models import User, UserResponse


class Thread(BaseModel):
    id: Optional[int] = None
    externalId: Optional[str] = None
    projectId: Optional[int] = None
    projectName: Optional[str] = None
    messages: List[Message]
    metadata: dict[str, Union[str, int]] = {}
    externalUser: Optional[User] = None
    createdAt: Optional[datetime] = None

class ThreadsQueryParams(BaseModel):
  class Config:
    extra = Extra.forbid

  pageSize: int = 50
  pageIndex: int = 0
  projectId: Optional[int] = None
  ids: Optional[List[int]] = None
  externalIds: Optional[List[str]] = None
  userSegmentIds: Optional[List[int]] = None
  issueIds: Optional[List[int]] = None
  intentIds: Optional[List[int]] = None
  search: Optional[str] = None
  hasFeedback: Optional[bool] = None
  feedbackType: Optional[Literal['POSITIVE', 'NEGATIVE']]= None
  attributeOptionIds: Optional[List[int]] = None
  outcome: Optional[Literal['successful', 'partial_success', 'unsuccessful']] = None
  metadataField: Optional[str] = None
  metadataValue: Optional[str] = None
  filterInternal: Optional[bool] = None
  filterAnonymousSessions: Optional[bool] = None
  includeFeedback: Optional[bool] = None
  includeIntents: Optional[bool] = None
  includeIssues: Optional[bool] = None
  before: Optional[datetime] = None
  after: Optional[datetime] = None

class SimpleProject(BaseModel):
    id: int
    name: str

class ThreadResponse(BaseModel):
    id: int
    organizationId: int
    externalId: Optional[str] = None

    project: SimpleProject

    externalUser: Optional[UserResponse] = None

    messages: List[MessageResponse]

    outcome: Optional[str] = None
    metadata: dict[str, Union[str, int]] = {}

    createdAt: datetime
    updatedAt: datetime

class ThreadsPagedResponse(BaseModel):
    count: int
    rows: List[ThreadResponse]



"""
Melodi: OpenAI API Wrapper with Full Logging

Melodi provides a seamless drop-in replacement for the OpenAI Python SDK,
adding automatic logging with minimal code changes. Simply update your import:

```diff
- import openai
+ from melodi.openai import openai
```
"""

import logging
import os
from typing import Optional, Callable

from packaging.version import Version
from wrapt import wrap_function_wrapper

from melodi.melodi_client import MelodiClient
from melodi.utils.openai_nonstream_extractor import (
    create_melodi_thread_from_openai_response,
)
from melodi.utils.openai_prompt_parser import _get_melodi_messages_from_openai_prompt
from melodi.utils.openai_stream_generator import (
    MelodiResponseGeneratorSync,
    MelodiResponseGeneratorAsync,
)
from melodi.utils.openai_utils import (
    OPENAI_CLIENTS_V0,
    OPENAI_CLIENTS_V1,
    OpenAiDefinition,
    OpenAiKwargsExtractor,
    _is_openai_v1,
    _is_streaming_response,
)
from melodi.utils.utils import create_error_melodi_thread

try:
    import openai
except ImportError:
    raise ModuleNotFoundError("OpenAI not installed, please run: 'pip install openai'")

try:
    from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI  # noqa: F401
except ImportError:
    AsyncAzureOpenAI = None
    AsyncOpenAI = None
    AzureOpenAI = None
    OpenAI = None

logger = logging.getLogger("melodi")


def melodi_openai_wrapper(func):
    def melodi_wrapper(open_ai_definitions, initialize):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                open_ai_definitions,
                initialize,
                wrapped,
                kwargs,
            )

        return wrapper

    return melodi_wrapper


@melodi_openai_wrapper
def _wrap(
    openai_resource: OpenAiDefinition,
    melodi_initialize_func: Callable,
    wrapped: Callable,
    kwargs: dict,
):
    melodi_client = melodi_initialize_func()
    arg_extractor = OpenAiKwargsExtractor(**kwargs)

    prompt_messages = _get_melodi_messages_from_openai_prompt(kwargs, openai_resource)
    try:
        openai_response = wrapped(**arg_extractor.get_openai_args())

        if _is_streaming_response(openai_response):
            try:
                return MelodiResponseGeneratorSync(
                    openai_resource=openai_resource,
                    openai_response=openai_response,
                    melodi_client=melodi_client,
                    prompt_messages=prompt_messages,
                )
            except Exception as ex:
                logger.error(f"Could not create Melodi thread out of streamed response: {repr(ex)}")
        else:
            create_melodi_thread_from_openai_response(
                openai_resource=openai_resource,
                openai_response=openai_response,
                prompt_messages=prompt_messages,
                melodi_client=melodi_client,
            )

        return openai_response
    except Exception as ex:
        logger.warning(ex)
        create_error_melodi_thread(
            melodi_client=melodi_client,
            prompt_messages=prompt_messages,
            model=kwargs.get("model"),
            exception=str(ex),
        )
        raise ex


@melodi_openai_wrapper
async def _wrap_async(
    openai_resource: OpenAiDefinition,
    melodi_initialize_func: Callable,
    wrapped: Callable,
    kwargs: dict,
):
    melodi_client = melodi_initialize_func()
    arg_extractor = OpenAiKwargsExtractor(**kwargs)

    prompt_messages = _get_melodi_messages_from_openai_prompt(kwargs, openai_resource)

    try:
        openai_response = await wrapped(**arg_extractor.get_openai_args())

        if _is_streaming_response(openai_response):
            try:
                return MelodiResponseGeneratorAsync(
                    openai_resource=openai_resource,
                    openai_response=openai_response,
                    melodi_client=melodi_client,
                    prompt_messages=prompt_messages,
                )
            except Exception as ex:
                logger.error(f"Could not create Melodi thread out of streamed response: {repr(ex)}")
        else:
            create_melodi_thread_from_openai_response(
                openai_resource=openai_resource,
                openai_response=openai_response,
                prompt_messages=prompt_messages,
                melodi_client=melodi_client,
            )

        return openai_response
    except Exception as ex:
        logger.warning(ex)
        create_error_melodi_thread(
            melodi_client=melodi_client,
            prompt_messages=prompt_messages,
            model=kwargs.get("model"),
            exception=str(ex),
        )
        raise ex


class OpenAIMelodi:
    melodi_client: Optional[MelodiClient] = None

    def initialize(self):
        if self.melodi_client is None:
            self.melodi_client = MelodiClient(
                api_key=os.getenv("MELODI_API_KEY"), verbose=True
            )

        return self.melodi_client

    def register_tracing(self):
        resources = OPENAI_CLIENTS_V1 if _is_openai_v1() else OPENAI_CLIENTS_V0

        for resource in resources:
            if resource.min_version is not None and Version(
                openai.__version__
            ) < Version(resource.min_version):
                continue

            wrap_function_wrapper(
                module=resource.module,
                name=f"{resource.object}.{resource.method}",
                wrapper=_wrap(
                    open_ai_definitions=resource,
                    initialize=self.initialize,
                )
                if resource.sync
                else _wrap_async(
                    open_ai_definitions=resource,
                    initialize=self.initialize,
                ),
            )


modifier = OpenAIMelodi()
modifier.register_tracing()



import logging
import os

from melodi.threads.data_models import Thread
from melodi.utils.openai_utils import time_now

logger = logging.getLogger("melodi")


def handle_melodi_failure(value):
    def decorate(wrapped_func):
        def applicator(*args, **kwargs):
            try:
                return wrapped_func(*args, **kwargs)
            except Exception as e:
                logger.error(f"{value}: {repr(e)}")
                return

        return applicator

    return decorate


@handle_melodi_failure("Could not create a Melodi thread")
def create_melodi_thread(melodi_client, melodi_messages, response_id, prompt_messages):
    logger.info("Creating Melodi thread ...")
    thread_metadata = {
        "created": time_now(as_string=True),
        "response_id": response_id,
    }
    # Create Melodi thread
    thread = Thread(
        projectId=os.getenv("MELODI_PROJECT_ID"),
        externalId=response_id,
        messages=prompt_messages + melodi_messages,
        metadata=thread_metadata,
    )
    melodi_client.threads.create(thread)
    logger.info("Done creating Melodi thread.")


@handle_melodi_failure("Could not create a Melodi thread")
def create_error_melodi_thread(melodi_client, prompt_messages, model, exception):
    logger.warning("Creating Melodi error thread ...")
    metadata = {
        "completion_tokens": 0,
        "prompt_tokens": 0,
        "total_tokens": 0,
        "openai_error": exception,
        "created": time_now(as_string=True),
    }
    if model:
        metadata["model"] = model
    melodi_error_thread = Thread(
        projectId=os.getenv("MELODI_PROJECT_ID"),
        messages=prompt_messages,
        metadata=metadata,
    )
    melodi_client.threads.create(melodi_error_thread)
    logger.warning("Done creating Melodi error thread.")



from melodi.messages.data_models import Message
from melodi.utils.openai_utils import (
    OpenAiDefinition,
    _is_openai_v1,
    COMPLETION_CHOICE_MESSAGE_KEYS,
    COMPLETION_CHOICE_KEYS,
    COMPLETION_USAGE_KEYS,
    COMPLETION_USAGE_TOKENS_KEYS,
    COMPLETION_USAGE_PROMPT_TOKENS_KEYS,
    NON_STREAM_MESSAGE_KEYS, parse_metadata_value,
)
from melodi.utils.utils import create_melodi_thread, handle_melodi_failure


@handle_melodi_failure("Could not create a Melodi thread")
def create_melodi_thread_from_openai_response(
    openai_resource: OpenAiDefinition,
    openai_response,
    melodi_client,
    prompt_messages: list,
):
    melodi_messages, response_id = _get_melodi_messages_from_openai_response(
        resource=openai_resource,
        response=openai_response
    )
    create_melodi_thread(
        melodi_client=melodi_client,
        melodi_messages=melodi_messages,
        response_id=response_id,
        prompt_messages=prompt_messages,
    )


def _get_melodi_messages_from_openai_response(resource: OpenAiDefinition, response):
    if response is None:
        return [], None

    response = to_dict(response)

    response_contents = []
    if resource.type == "completion":
        choices = response.get("choices", [])
        if len(choices) > 0:
            choice = choices[-1]
            response_contents = (
                [_extract_chat_base_response(choice.text)]
                if _is_openai_v1()
                else [_extract_chat_base_response(choice.get("text"))]
            )
    elif resource.type == "chat":
        choices = response.get("choices", [])
        # If multiple choices were generated, we'll show all of them in the UI as a list.
        if len(choices) > 0:
            response_contents = [
                _extract_chat_response(to_dict(choice))
                if _is_openai_v1()
                else _extract_chat_base_response(choice.get("message"))
                for choice in choices
            ]

    response_metadata = _get_response_metadata(response)
    melodi_messages = []
    for message_dict in response_contents:
        if len(response_contents) > 1:
            id_suffix = f"-alternative-{len(melodi_messages)}"
        else:
            id_suffix = ""

        content = message_dict.pop("content", None)
        metadata = {"type": "response"}
        metadata.update(**message_dict)
        metadata.update(**response_metadata)

        melodi_messages.append(
            Message(
                externalId=f'{metadata.get("id")}{id_suffix}',
                role=metadata.get("role").title(),
                content=content,
                metadata=metadata,
            )
        )
    return melodi_messages, response.get("id")


def _extract_chat_base_response(content: str):
    """Extracts the content from the response."""
    return {
        "content": content,
    }


def _extract_chat_response(choice_dict: dict):
    """Extracts the message content from an OpenAI choice."""
    message_dict = to_dict(choice_dict.get("message", {}))
    if not message_dict:
        return {}

    response = {}
    for message_key in COMPLETION_CHOICE_MESSAGE_KEYS:
        value = message_dict.get(message_key)
        if value is None:
            continue

        response[message_key] = parse_metadata_value(value)

    for choice_key in COMPLETION_CHOICE_KEYS:
        value = choice_dict.get(choice_key)
        if value is None:
            continue

        response[choice_key] = parse_metadata_value(value)
    return response


def _get_response_metadata(response):
    usage_dict = to_dict(response.get("usage", {}))
    completion_tokens_dict = to_dict(usage_dict.get("completion_tokens_details", {}))
    prompt_tokens_dict = to_dict(usage_dict.get("prompt_tokens_details", {}))

    metadata = {}
    for key in NON_STREAM_MESSAGE_KEYS:
        value = response.get(key)
        if value is None:
            continue

        metadata[key] = parse_metadata_value(value)

    for key in COMPLETION_USAGE_KEYS:
        value = usage_dict.get(key)
        if value is None:
            continue

        metadata[key] = parse_metadata_value(value)

    for key in COMPLETION_USAGE_TOKENS_KEYS:
        value = completion_tokens_dict.get(key)
        if value is None:
            continue

        metadata[key] = parse_metadata_value(value)

    for key in COMPLETION_USAGE_PROMPT_TOKENS_KEYS:
        value = prompt_tokens_dict.get(key)
        if not value:
            continue

        metadata[key] = parse_metadata_value(value)

    return metadata


def to_dict(input_value):
    if input_value is None:
        return None

    if isinstance(input_value, dict):
        return input_value

    return input_value.__dict__



from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel

from melodi.feedback.data_models import FeedbackResponse
from melodi.intents.data_models import IntentResponse
from melodi.issues.data_models import IssueResponse


class IntentMessageAssociation(BaseModel):
    id: int
    intentId: int
    messageId: int
    userId: Optional[int] = None
    intent: IntentResponse

class IssueMessageAssociation(BaseModel):
    id: int
    issueId: int
    messageId: int
    userId: Optional[int] = None
    issue: IssueResponse

class Message(BaseModel):
    externalId: Optional[str] = None
    type: Literal['markdown', 'json'] = 'markdown'
    role: str
    content: Optional[str] = None
    jsonContent: Optional[Any] = None
    metadata: dict[str, Union[str, int]] = {}

class MessageResponse(Message):
    id: int
    issueAssociations: List[IssueMessageAssociation] = []
    intentAssociations: List[IntentMessageAssociation] = []
    externalFeedback: List[FeedbackResponse] = []



import logging

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.logging import _log_melodi_http_errors
from melodi.users.data_models import User, UserResponse, UsersPagedResponse, UsersQueryParams


class UserClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = (
            self.base_url + "/api/external/users"
        )
        self.endpoint = (
            self.base_endpoint + f"?apiKey={self.api_key}"
        )

        self.logger = logging.getLogger(__name__)

    def get(self, query_params: UsersQueryParams = UsersQueryParams()) -> UsersPagedResponse:
        url = f"{self.endpoint}&pageIndex={query_params.pageIndex}&pageSize={query_params.pageSize}"

        if (query_params.projectId):
            url = f"{url}&projectId={query_params.projectId}"
        if (query_params.before):
            url = f"{url}&before={query_params.before.isoformat()}"
        if (query_params.after):
            url = f"{url}&after={query_params.after.isoformat()}"

        try:
            response = requests.request("GET", url)

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()

            return parse_obj_as(UsersPagedResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def create_or_update(self, user: User) -> UserResponse:
        url = f"{self.base_endpoint}?apiKey={self.api_key}"

        try:
            response = requests.put(
                url, headers=self._get_headers(), json=user.dict(by_alias=True)
            )

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(UserResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)

    def update(self, user: User) -> User:
        url = f"{self.base_endpoint}/{user.externalId}?apiKey={self.api_key}"

        try:
            response = requests.patch(
                url, headers=self._get_headers(), json=user.dict(by_alias=True)
            )

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(UserResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)



from datetime import datetime
from typing import List, Literal, Optional

from pydantic import BaseModel

from melodi.users.data_models import User, UserResponse


class BaseFeedback(BaseModel):
    feedbackType: Optional[Literal['POSITIVE', 'NEGATIVE']] = None
    feedbackText: Optional[str] = None

class Feedback(BaseFeedback):
    projectId: Optional[int] = None

    externalThreadId: Optional[str] = None
    externalMessageId: Optional[str] = None

    externalUser: Optional[User] = None

    attributes: Optional[dict[str, str]] = {}
    createdAt: Optional[datetime] = None

class Attribute(BaseModel):
    id: int
    name: str

class AttributeOption(BaseModel):
    id: int
    name: str
    attribute: Attribute

class FeedbackResponse(BaseFeedback):
    id: int
    projectId: int

    externalUserId: Optional[int] = None
    externalUser: Optional[UserResponse] = None

    attributeOptions: List[AttributeOption] = []
    createdAt: datetime
    updatedAt: datetime

class FeedbackCreateOrUpdateRequest(Feedback):
    id: Optional[int] = None


from typing import List

from pydantic import BaseModel, Extra


class BulkUserInternalForProjectRequest(BaseModel):
  class Config:
    extra = Extra.forbid

  projectId: int
  externalUserIds: List[int]


class BulkUserInternalForProjectResponse(BaseModel):
  count: int



from collections import defaultdict

from melodi.melodi_client import MelodiClient
from melodi.messages.data_models import Message
from melodi.utils.openai_nonstream_extractor import to_dict
from melodi.utils.openai_utils import (
    OpenAiDefinition,
    clean_dict_value,
    _is_openai_v1,
    parse_metadata_value,
)
from melodi.utils.utils import create_melodi_thread


def _extract_streamed_openai_response(resource: OpenAiDefinition, chunks: list):
    if not chunks:
        return None, None

    if resource.type == "chat":
        completion = defaultdict(str)
    else:
        completion = {"role": "assistant", "content": ""}
    response_id, model, usage, finish_reason, created_at = None, None, None, None, None
    for chunk in chunks:
        if _is_openai_v1():
            chunk = chunk.__dict__

        model = chunk.get("model")
        response_id = chunk.get("id")
        usage = to_dict(chunk.get("usage"))
        created_at = chunk.get("created")

        choices = chunk.get("choices", [])
        for choice in choices:
            if _is_openai_v1():
                choice = choice.__dict__

            if resource.type == "chat":
                delta = choice.get("delta")

                # The finish_reason is included only once, in the final one
                finish_reason = choice.get("finish_reason")

                if _is_openai_v1():
                    delta = delta.__dict__

                # The role is included only once
                if delta.get("role") is not None:
                    completion["role"] = delta["role"]

                # The full content is made out of all the chunk contents
                if delta.get("content") is not None:
                    completion["content"] = (
                        delta.get("content")
                        if completion["content"] is None
                        else completion["content"] + delta.get("content")
                    )
                # TODO this is deprecated
                elif delta.get("function_call") is not None:
                    curr = completion["function_call"]
                    function_call_chunk = delta.get("function_call")

                    if not curr:
                        completion["function_call"] = {
                            "name": getattr(function_call_chunk, "name", ""),
                            "arguments": getattr(function_call_chunk, "arguments", ""),
                        }

                    else:
                        curr["name"] = curr["name"] or getattr(
                            function_call_chunk, "name", None
                        )
                        curr["arguments"] += getattr(
                            function_call_chunk, "arguments", ""
                        )
                elif delta.get("tool_calls") is not None:
                    curr = completion["tool_calls"]
                    tool_call_chunk = getattr(
                        delta.get("tool_calls")[0], "function", None
                    )

                    if not curr:
                        completion["tool_calls"] = [
                            {
                                "name": getattr(tool_call_chunk, "name", ""),
                                "arguments": getattr(tool_call_chunk, "arguments", ""),
                            }
                        ]
                    elif getattr(tool_call_chunk, "name") is not None:
                        curr.append(
                            {
                                "name": getattr(tool_call_chunk, "name", None),
                                "arguments": getattr(
                                    tool_call_chunk, "arguments", None
                                ),
                            }
                        )
                    else:
                        curr[-1]["name"] = curr[-1]["name"] or getattr(
                            tool_call_chunk, "name", None
                        )
                        curr[-1]["arguments"] += getattr(
                            tool_call_chunk, "arguments", None
                        )

            if resource.type == "completion":
                completion["content"] += choice.get("text", "")

    message_metadata = {
        "model": model,
        "usage": usage,
        "finish_reason": finish_reason,
        "created_at": created_at,
        "tool_calls": parse_metadata_value(completion.get("tool_calls")),
    }

    melodi_message = Message(
        externalId=response_id,
        role=completion["role"].title() if completion["role"] else None,
        content=completion["content"] if completion["role"] else None,
        metadata=clean_dict_value(message_metadata),
    )

    return melodi_message, response_id


class MelodiResponseGeneratorSync:
    def __init__(
        self,
        *,
        openai_resource: OpenAiDefinition,
        openai_response,
        melodi_client: MelodiClient,
        prompt_messages: list,
    ):
        self.items = []

        self.openai_resource = openai_resource
        self.openai_response = openai_response
        self.melodi_client = melodi_client
        self.prompt_messages = prompt_messages

    def __iter__(self):
        try:
            for i in self.openai_response:
                self.items.append(i)

                yield i
        finally:
            self._finalize()

    def __next__(self):
        try:
            item = self.openai_response.__next__()
            self.items.append(item)

            return item

        except StopIteration:
            self._finalize()

            raise

    def __enter__(self):
        return self.__iter__()

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _finalize(self):
        melodi_message, response_id = _extract_streamed_openai_response(
            self.openai_resource, self.items
        )

        create_melodi_thread(
            melodi_client=self.melodi_client,
            melodi_messages=[melodi_message],
            response_id=response_id,
            prompt_messages=self.prompt_messages,
        )


class MelodiResponseGeneratorAsync:
    def __init__(
        self,
        *,
        openai_resource: OpenAiDefinition,
        openai_response,
        melodi_client: MelodiClient,
        prompt_messages: list,
    ):
        self.items = []

        self.openai_resource = openai_resource
        self.openai_response = openai_response
        self.melodi_client = melodi_client
        self.prompt_messages = prompt_messages

    async def __aiter__(self):
        try:
            async for i in self.openai_response:
                self.items.append(i)

                yield i
        finally:
            await self._finalize()

    async def __anext__(self):
        try:
            item = await self.openai_response.__anext__()
            self.items.append(item)

            return item

        except StopAsyncIteration:
            await self._finalize()

            raise

    async def __aenter__(self):
        return self.__aiter__()

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    async def _finalize(self):
        melodi_message, response_id = _extract_streamed_openai_response(
            self.openai_resource, self.items
        )

        create_melodi_thread(
            melodi_client=self.melodi_client,
            melodi_messages=[melodi_message],
            response_id=response_id,
            prompt_messages=self.prompt_messages,
        )

    async def close(self) -> None:
        """Close the response and release the connection.

        Automatically called if the response body is read to completion.
        """
        await self.openai_response.close()

    async def aclose(self) -> None:
        """Close the response and release the connection.

        Automatically called if the response body is read to completion.
        """
        await self.openai_response.aclose()



from datetime import datetime

from pydantic import BaseModel


class IntentResponse(BaseModel):
    id: int
    name: str
    createdAt: datetime


class IntentUpsertRequest(BaseModel):
    projectId: int
    name: str
    description: str
    createdAt: datetime





import unittest
from dataclasses import dataclass
from typing import Optional
from unittest.mock import patch, MagicMock

from melodi.utils.openai_stream_generator import (
    _extract_streamed_openai_response,
    MelodiResponseGeneratorSync,
)
from melodi.utils.openai_utils import OpenAiDefinition


@dataclass
class ChoiceDeltaToolCallFunction:
    """Testing class which mirrors OpenAI's objects."""

    arguments: str
    name: Optional[str]


@dataclass
class ChoiceDeltaToolCall:
    """Testing class which mirrors OpenAI's objects."""

    index: int
    id: Optional[str]
    function: Optional[ChoiceDeltaToolCallFunction]
    index: Optional[int]
    type: Optional[str]
    logprobs: Optional[list] = None
    content_filter_results: Optional[dict] = None
    finish_reason: Optional[str] = None


@dataclass
class ChoiceDelta:
    """Testing class which mirrors OpenAI's objects."""

    content: Optional[str]
    function_call: Optional[str]
    refusal: Optional[str]
    role: Optional[str]
    tool_calls: Optional[list]


@dataclass
class Choice:
    """Testing class which mirrors OpenAI's objects."""

    delta: ChoiceDelta
    finish_reason: Optional[str]
    index: int
    logprobs: Optional[list]
    content_filter_results: dict


@dataclass
class ChatCompletionChunk:
    """Testing class which mirrors OpenAI's objects."""

    id: str
    choices: Optional[list]
    created: int
    model: Optional[str]
    object: Optional[str]
    service_tier: Optional[str]
    system_fingerprint: Optional[str]
    usage: Optional[dict]
    prompt_filter_results: Optional[list] = None


class TestOpenAIStreamExtractorTests(unittest.TestCase):

    openai_mock_response_function_call = [
        ChatCompletionChunk(
            id="",
            choices=[],
            created=0,
            model="",
            object="",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
            prompt_filter_results=[
                {
                    "prompt_index": 0,
                    "content_filter_results": {
                        "hate": {"filtered": False, "severity": "safe"},
                        "jailbreak": {"filtered": False, "detected": False},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                }
            ],
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role="assistant",
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id="call_nuMz0OzAFocR9il88r5794xN",
                                function=ChoiceDeltaToolCallFunction(
                                    arguments="", name="get_current_weather"
                                ),
                                type="function",
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=None,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments='{"', name=None
                                ),
                                type=None,
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=None,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments="location", name=None
                                ),
                                type=None,
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=None,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments='":"', name=None
                                ),
                                type=None,
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=None,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments="Paris", name=None
                                ),
                                type=None,
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=[
                            ChoiceDeltaToolCall(
                                index=0,
                                id=None,
                                function=ChoiceDeltaToolCallFunction(
                                    arguments='"}', name=None
                                ),
                                type=None,
                            )
                        ],
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason="tool_calls",
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748289435,
            model="gpt-4.1-2025-04-14",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint="fp_07e970ab25",
            usage=None,
        ),
    ]

    openai_mock_response = [
        ChatCompletionChunk(
            id="",
            choices=[],
            created=0,
            model="",
            object="",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
            prompt_filter_results=[
                {
                    "prompt_index": 0,
                    "content_filter_results": {
                        "hate": {"filtered": False, "severity": "safe"},
                        "jailbreak": {"filtered": False, "detected": False},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                }
            ],
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content="",
                        function_call=None,
                        refusal=None,
                        role="assistant",
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content="this",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" is",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" meant",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" to",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" be",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" a",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" streamed",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=" response",
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason=None,
                    index=0,
                    logprobs=None,
                    content_filter_results={
                        "hate": {"filtered": False, "severity": "safe"},
                        "self_harm": {"filtered": False, "severity": "safe"},
                        "sexual": {"filtered": False, "severity": "safe"},
                        "violence": {"filtered": False, "severity": "safe"},
                    },
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
        ChatCompletionChunk(
            id="chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
            choices=[
                Choice(
                    delta=ChoiceDelta(
                        content=None,
                        function_call=None,
                        refusal=None,
                        role=None,
                        tool_calls=None,
                    ),
                    finish_reason="stop",
                    index=0,
                    logprobs=None,
                    content_filter_results={},
                )
            ],
            created=1748283610,
            model="o4-mini-2025-04-16",
            object="chat.completion.chunk",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        ),
    ]

    def test_extract_streamed_openai_response(self):
        melodi_message, response_id = _extract_streamed_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            chunks=[],
        )
        self.assertIsNone(melodi_message)
        self.assertIsNone(response_id)

        melodi_message, response_id = _extract_streamed_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            chunks=self.openai_mock_response,
        )

        self.assertEqual(
            melodi_message.externalId, "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz"
        )
        self.assertEqual(melodi_message.role, "Assistant")
        self.assertEqual(
            melodi_message.content, "this is meant to be a streamed response"
        )
        self.assertEqual(
            melodi_message.metadata,
            {
                "model": "o4-mini-2025-04-16",
                "finish_reason": "stop",
                "created_at": 1748283610,
            },
        )
        self.assertEqual(response_id, "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz")

    def test_extract_streamed_openai_response_function_tools(self):
        melodi_message, response_id = _extract_streamed_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            chunks=self.openai_mock_response_function_call,
        )

        self.assertEqual(
            melodi_message.externalId, "chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR"
        )
        self.assertEqual(melodi_message.role, "Assistant")
        self.assertEqual(melodi_message.content, "")
        self.assertEqual(
            melodi_message.metadata,
            {
                "model": "gpt-4.1-2025-04-14",
                "finish_reason": "tool_calls",
                "created_at": 1748289435,
                "tool_calls": '[{"name": "get_current_weather", "arguments": "{\\"location\\":\\"Paris\\"}"}]',
            },
        )
        self.assertEqual(response_id, "chatcmpl-BbYEdy7cBIbJsbLXouL7Z47t93LUR")

    @patch("melodi.utils.openai_stream_generator.create_melodi_thread")
    def test_melodi_response_generator_sync(self, mock_melodi_thread):
        melodi_client_mock = MagicMock()
        result = MelodiResponseGeneratorSync(
            openai_resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            openai_response=[],
            melodi_client=melodi_client_mock,
            prompt_messages=[],
        )
        self.assertEqual(len(mock_melodi_thread.mock_calls), 0)

        for _ in result:
            pass
        self.assertEqual(len(mock_melodi_thread.mock_calls), 1)
        _, _, kwargs = mock_melodi_thread.mock_calls[0]
        self.assertEqual(kwargs["melodi_client"], melodi_client_mock)
        self.assertEqual(kwargs["melodi_messages"], [None])
        self.assertEqual(kwargs["response_id"], None)
        self.assertEqual(kwargs["prompt_messages"], [])

        result = MelodiResponseGeneratorSync(
            openai_resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            openai_response=self.openai_mock_response,
            melodi_client=melodi_client_mock,
            prompt_messages=[],
        )
        for _ in result:
            pass
        self.assertEqual(len(mock_melodi_thread.mock_calls), 2)
        _, _, kwargs = mock_melodi_thread.mock_calls[1]
        self.assertEqual(kwargs["melodi_client"], melodi_client_mock)
        self.assertEqual(len(kwargs["melodi_messages"]), 1)
        self.assertEqual(
            kwargs["melodi_messages"][0].externalId,
            "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
        )
        self.assertEqual(kwargs["melodi_messages"][0].type, "markdown")
        self.assertEqual(kwargs["melodi_messages"][0].role, "Assistant")
        self.assertEqual(
            kwargs["melodi_messages"][0].content,
            "this is meant to be a streamed response",
        )
        self.assertEqual(kwargs["melodi_messages"][0].jsonContent, None)
        self.assertEqual(
            kwargs["melodi_messages"][0].metadata,
            {
                "model": "o4-mini-2025-04-16",
                "finish_reason": "stop",
                "created_at": 1748283610,
            },
        )
        self.assertEqual(
            kwargs["response_id"], "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz"
        )
        self.assertEqual(kwargs["prompt_messages"], [])

    @patch("melodi.utils.openai_stream_generator.create_melodi_thread")
    def test_melodi_response_generator_async(self, mock_melodi_thread):
        melodi_client_mock = MagicMock()
        result = MelodiResponseGeneratorSync(
            openai_resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=False,
            ),
            openai_response=[],
            melodi_client=melodi_client_mock,
            prompt_messages=[],
        )
        self.assertEqual(len(mock_melodi_thread.mock_calls), 0)

        for _ in result:
            pass
        self.assertEqual(len(mock_melodi_thread.mock_calls), 1)
        _, _, kwargs = mock_melodi_thread.mock_calls[0]
        self.assertEqual(kwargs["melodi_client"], melodi_client_mock)
        self.assertEqual(kwargs["melodi_messages"], [None])
        self.assertEqual(kwargs["response_id"], None)
        self.assertEqual(kwargs["prompt_messages"], [])

        result = MelodiResponseGeneratorSync(
            openai_resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=False,
            ),
            openai_response=self.openai_mock_response,
            melodi_client=melodi_client_mock,
            prompt_messages=[],
        )
        for _ in result:
            pass
        self.assertEqual(len(mock_melodi_thread.mock_calls), 2)
        _, _, kwargs = mock_melodi_thread.mock_calls[1]
        self.assertEqual(kwargs["melodi_client"], melodi_client_mock)
        self.assertEqual(len(kwargs["melodi_messages"]), 1)
        self.assertEqual(
            kwargs["melodi_messages"][0].externalId,
            "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz",
        )
        self.assertEqual(kwargs["melodi_messages"][0].type, "markdown")
        self.assertEqual(kwargs["melodi_messages"][0].role, "Assistant")
        self.assertEqual(
            kwargs["melodi_messages"][0].content,
            "this is meant to be a streamed response",
        )
        self.assertEqual(kwargs["melodi_messages"][0].jsonContent, None)
        self.assertEqual(
            kwargs["melodi_messages"][0].metadata,
            {
                "model": "o4-mini-2025-04-16",
                "finish_reason": "stop",
                "created_at": 1748283610,
            },
        )
        self.assertEqual(
            kwargs["response_id"], "chatcmpl-BbWigtG2AjFsEEXkAzbBlmiAL8sfz"
        )
        self.assertEqual(kwargs["prompt_messages"], [])


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

from melodi.messages.data_models import Message
from melodi.utils.openai_nonstream_extractor import (
    _get_response_metadata,
    _extract_chat_base_response,
    _get_melodi_messages_from_openai_response, create_melodi_thread_from_openai_response,
)
from melodi.utils.openai_utils import OpenAiDefinition


class TestOpenAINonstreamExtractorTests(unittest.TestCase):
    def test_get_response_metadata(self):
        self.assertEqual(_get_response_metadata({}), {})
        self.assertEqual(
            _get_response_metadata(
                {
                    "usage": {
                        "completion_tokens_details": {
                            "accepted_prediction_tokens": 12,
                            "reasoning_tokens": 34,
                            "rejected_prediction_tokens": 2,
                        },
                        "prompt_tokens_details": {
                            "cached_tokens": 1,
                        },
                        "total_tokens": 67,
                    },
                    "created": 1234,
                    "id": 1111,
                    "model": "openai-mini-v12",
                    "object": "text",
                    "service_tier": "pro",
                    "system_fingerprint": 123,
                }
            ),
            {
                "accepted_prediction_tokens": 12,
                "reasoning_tokens": 34,
                "rejected_prediction_tokens": 2,
                "cached_tokens": 1,
                "total_tokens": 67,
                "created": 1234,
                "id": 1111,
                "model": "openai-mini-v12",
                "object": "text",
                "service_tier": "pro",
                "system_fingerprint": 123,
            },
        )
        self.assertEqual(
            _get_response_metadata(
                {
                    "usage": {
                        "completion_tokens_details": {
                            "reasoning_tokens": 34,
                            "rejected_prediction_tokens": 0,
                        },
                        "prompt_tokens_details": {
                            "cached_tokens": 1,
                        },
                        "total_tokens": 67,
                    },
                    "created": 1234,
                    "id": 1111,
                    "model": "openai-mini-v12",
                    "object": "text",
                    "service_tier": "pro",
                    "system_fingerprint": 123,
                }
            ),
            {
                "reasoning_tokens": 34,
                "rejected_prediction_tokens": 0,
                "cached_tokens": 1,
                "total_tokens": 67,
                "created": 1234,
                "id": 1111,
                "model": "openai-mini-v12",
                "object": "text",
                "service_tier": "pro",
                "system_fingerprint": 123,
            },
        )

    def test_extract_chat_base_response(self):
        self.assertEqual(_extract_chat_base_response("test"), {"content": "test"})

    def test_get_melodi_messages_from_openai_response(self):
        messages, message_id = _get_melodi_messages_from_openai_response(None, None)
        self.assertEqual(messages, [])
        self.assertEqual(message_id, None)

        messages, message_id = _get_melodi_messages_from_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            response=None,
        )
        self.assertEqual(messages, [])
        self.assertEqual(message_id, None)

        messages, message_id = _get_melodi_messages_from_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            response={},
        )
        self.assertEqual(messages, [])
        self.assertEqual(message_id, None)

        messages, message_id = _get_melodi_messages_from_openai_response(
            resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            response={
                "choices": [
                    {
                        "message": {
                            "content": "Hi",
                            "refusal": None,
                            "role": "new role",
                        },
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": [],
                    }
                ],
                "id": "message_1",
            },
        )
        self.assertEqual(
            messages,
            [
                Message(
                    externalId="message_1",
                    role="New Role",
                    content="Hi",
                    jsonContent=None,
                    metadata={
                        "id": "message_1",
                        "type": "response",
                        "role": "new role",
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": "[]",
                    },
                )
            ],
        )
        self.assertEqual(message_id, "message_1")

    @patch("melodi.utils.openai_nonstream_extractor.create_melodi_thread")
    def test_create_melodi_thread_from_openai_response(self, melodi_patch):
        create_melodi_thread_from_openai_response(
            openai_resource=None,
            openai_response=None,
            melodi_client=None,
            prompt_messages=[]
        )
        _, _, kwargs = melodi_patch.mock_calls[0]
        self.assertEqual(
            kwargs,
            {
                "melodi_client": None,
                "melodi_messages": [],
                "response_id": None,
                "prompt_messages": [],
            }
        )

        create_melodi_thread_from_openai_response(
            openai_resource=OpenAiDefinition(
                module="openai",
                object="ChatCompletion",
                method="create",
                type="chat",
                sync=True,
            ),
            openai_response={
                "choices": [
                    {
                        "message": {
                            "content": "Hi",
                            "refusal": None,
                            "role": "new role",
                        },
                        "finish_reason": "stop",
                        "index": 0,
                        "logprobs": [],
                    }
                ],
                "id": "message_1",
            },
            melodi_client=None,
            prompt_messages=[]
        )
        _, _, kwargs = melodi_patch.mock_calls[1]
        self.assertEqual(
            kwargs,
            {
                "melodi_client": None,
                "melodi_messages": [
                    Message(
                        externalId="message_1",
                        role="New Role",
                        content="Hi",
                        jsonContent=None,
                        metadata={
                            "id": "message_1",
                            "type": "response",
                            "role": "new role",
                            "finish_reason": "stop",
                            "index": 0,
                            "logprobs": "[]",
                        },
                    )
                ],
                "response_id": "message_1",
                "prompt_messages": [],
            }
        )


if __name__ == "__main__":
    unittest.main()



import logging

import requests
from pydantic import parse_obj_as

from melodi.base_client import BaseClient
from melodi.exceptions import MelodiAPIError
from melodi.issues.data_models import IssueResponse, IssueUpsertRequest
from melodi.logging import _log_melodi_http_errors


class IssuesClient(BaseClient):
    def __init__(self, base_url: str, api_key: str):
        self.api_key = api_key
        self.base_url = base_url

        self.base_endpoint = self.base_url + "/api/external/issues"
        self.endpoint = self.base_endpoint + f"?apiKey={self.api_key}"

        self.logger = logging.getLogger(__name__)

    def upsert(self, issueUpsertRequest: IssueUpsertRequest) -> IssueResponse:
        try:
            response = requests.put(
                self.endpoint, headers=self._get_headers(), json=issueUpsertRequest.dict(by_alias=True)
            )

            _log_melodi_http_errors(self.logger, response)
            response.raise_for_status()
            return parse_obj_as(IssueResponse, response.json())
        except MelodiAPIError as e:
            raise MelodiAPIError(e)



import os
import unittest
from unittest import IsolatedAsyncioTestCase

from melodi.utils.openai import AsyncAzureOpenAI


class TestAsyncWrapper(IsolatedAsyncioTestCase):

    @unittest.skip
    async def test_openai_azure_async(self):
        client = AsyncAzureOpenAI(
            api_key=os.getenv("AZURE_API_KEY"),
            api_version="2024-12-01-preview",
            azure_endpoint=os.getenv("AZURE_ENDPOINT"),
        )

        response = await client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                },
                {
                    "role": "user",
                    "content": "I am going to Paris, what should I see?",
                }
            ],
            max_completion_tokens=5000,
            model="o4-mini"
        )

    @unittest.skip
    async def test_openai_azure_stream_response(self):
        client = AsyncAzureOpenAI(
            api_key=os.getenv("AZURE_API_KEY"),
            api_version="2025-03-01-preview",
            azure_endpoint=os.getenv("AZURE_ENDPOINT"),
        )

        stream = await client.chat.completions.create(
            model="o4-mini",
            messages=[
                {
                    "role": "user",
                    "content": "Say 'double bubble bath' 5 times fast.",
                },
            ],
            stream=True,
            n=1,
        )

        async for event in stream:
            # print(event)
            pass

    @unittest.skip
    async def test_azure_openai_streaming_function_tools(self):
        client = AsyncAzureOpenAI(
            api_key=os.getenv("AZURE_API_KEY"),
            api_version="2025-03-01-preview",
            azure_endpoint=os.getenv("AZURE_ENDPOINT"),
        )

        tools = [
            {
                "type": "function",
                "function": {
                    "name": "get_current_weather",
                    "description": "Get the current weather in a given location",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The city and state, e.g. San Francisco, CA",
                            },
                            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                        },
                        "required": ["location"],
                    },
                }
            }
        ]

        stream = await client.chat.completions.create(
            model="gpt-4.1",
            messages=[{"role": "user", "content": "What's the weather like in Brussels today?"}],
            tools=tools,
            stream=True
        )

        async for event in stream:
            pass

    @unittest.skip
    async def test_azure_openai_function_tools(self):
        client = AsyncAzureOpenAI(
            api_key=os.getenv("AZURE_API_KEY"),
            api_version="2024-12-01-preview",
            azure_endpoint=os.getenv("AZURE_ENDPOINT"),
        )

        tools = [
            {
                "type": "function",
                "function": {
                    "name": "get_current_weather",
                    "description": "Get the current weather in a given location",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "location": {
                                "type": "string",
                                "description": "The city and state, e.g. San Francisco, CA",
                            },
                            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                        },
                        "required": ["location"],
                    },
                }
            }
        ]
        messages = [{"role": "user", "content": "What's the weather like in Athens today?"}]
        completion = await client.chat.completions.create(
            model="gpt-4.1",
            messages=messages,
            tools=tools,
            tool_choice="auto"
        )


if __name__ == "__main__":
    unittest.main()



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Extra


class UserSegmentTypesQueryParams(BaseModel):
  class Config:
    extra = Extra.forbid

  projectId: Optional[int] = None

class UserSegmentDefinition(BaseModel):
    id: int
    name: str

    organizationId: int
    projectId: Optional[int] = None
    userSegmentTypeId: int

    createdAt: datetime
    updatedAt: datetime

class UserSegmentTypeDefinition(BaseModel):
    id: int
    name: str

    organizationId: int
    projectId: Optional[int] = None

    segments: List[UserSegmentDefinition]

    createdAt: datetime
    updatedAt: datetime
